
--
--  Overachiever
--    by Tuhljin
--


-- Overachiever_Debug = true;

local THIS_VERSION = GetAddOnMetadata("Overachiever", "Version")

Overachiever = {};

local CATEGORIES_ALL, CATEGORIES_EXPLOREZONES, CATEGORY_EXPLOREROOT
local OptionsPanel
local MadeDraggable_AchFrame, MadeDragSave_AchFrame

local AchievementIcon = "Interface\\AddOns\\Overachiever\\AchShield";

Overachiever.DefaultSettings = {
  Tooltip_ShowProgress = true;
  Tooltip_ShowProgress_Other = false;
  Tooltip_ShowID = false;
  Tracker_GreenCheck = true;
  Explore_AutoTrack = true;
  Explore_AutoTrack_Completed = false;
  CritterTip_loved = true;
  Draggable_AchFrame = true;
  DragSave_AchFrame = false;
  Version = THIS_VERSION;
};


local function chatprint(msg, premsg)
  premsg = premsg or "[Overachiever]"
  DEFAULT_CHAT_FRAME:AddMessage("|cff7eff00"..premsg.."|r "..msg, 0.741, 1, 0.467);
end

local function copytab(from, to)
  for k,v in pairs(from) do
    if(type(v) == "table") then
      to[k] = {}
      copytab(v, to[k]);
    else
      to[k] = v;
    end
  end
end

local function BuildCategoryInfo()
  CATEGORIES_ALL = GetCategoryList();
  local explorecat, name, parentID
  for i,id in ipairs(CATEGORIES_ALL) do
    name = GetCategoryInfo(id)
    if (name == OVERACHIEVER_STRINGS_CAT_EXPLORATION) then
      explorecat = id;
      break;
    end
  end
  assert(explorecat, "Exploration category not found.")
  CATEGORY_EXPLOREROOT = explorecat;
  CATEGORIES_EXPLOREZONES = {};
  for i,id in ipairs(CATEGORIES_ALL) do
    name, parentID = GetCategoryInfo(id)
    if (parentID == explorecat) then
      CATEGORIES_EXPLOREZONES[#(CATEGORIES_EXPLOREZONES) + 1] = id;
    end
  end
end

local function getSelectedAchievement(ignoreTab)
  if (AchievementFrame and (ignoreTab or AchievementFrame.selectedTab == 1)) then
    return AchievementFrameAchievements.selection;
  end
end

local function expandCategory(category)
-- Based on part of AchievementFrame_SelectAchievement() in Blizzard_AchievementUI.lua
  for i, entry in next, ACHIEVEMENTUI_CATEGORIES do
    if ( entry.id == category ) then
      entry.collapsed = false;
    elseif ( entry.parent == category ) then
      entry.hidden = false;
    end
  end
  AchievementFrameCategories_Update()
end

local function openToAchievement(id, canToggleTracking)
  assert( (type(id)=="number"), "Invalid achievement ID." )
  local sel
  if (not AchievementFrame or not AchievementFrame:IsShown()) then
    ToggleAchievementFrame()
  elseif (canToggleTracking) then
    sel = getSelectedAchievement()
  end
  if (sel == id) then
    AchievementButton_ToggleTracking(id)
  else
    AchievementFrame_SelectAchievement(id)
    local category = GetAchievementCategory(id)
    local _, parentID = GetCategoryInfo(category)
    if (parentID == -1) then
      expandCategory(category)
    end
  end
end

local function isExplorationAchievement(id, zonesOnly)
  local name, parentID = GetCategoryInfo( GetAchievementCategory(id) )
  if (parentID == -1) then
    if (zonesOnly) then  return false;  end
  else
    name = GetCategoryInfo(parentID);
  end
  if (name == OVERACHIEVER_STRINGS_CAT_EXPLORATION) then  return true;  end
end

local function getCategoryID(name)
  local n
  for i,id in ipairs(CATEGORIES_ALL) do
    n = GetCategoryInfo(id)
    if (n == name) then  return id;  end
  end
end

local searchResults_cat
local function getAchievementID_cat(category, name, anyCase, allowPartial, getAll)
  if (getAll) then
    if (searchResults_cat) then
      for k,v in pairs(searchResults_cat) do  -- Reuse table to avoid garbage generation
        searchResults_cat[k] = nil
      end
    else
      searchResults_cat = {}
    end
  end
  local id, n
  if (anyCase) then  name = strlower(name);  end
  for i=1,GetCategoryNumAchievements(category) do
    id, n = GetAchievementInfo(category, i)
    if (anyCase) then  n = strlower(n);  end
    if ( n == name or (allowPartial and strfind(n, name, 1, true)) ) then
      if (getAll) then
        searchResults_cat[#(searchResults_cat) + 1] = id;
      else
        return id;
      end
    end
  end
  if (getAll and #(searchResults_cat) > 0) then
    return searchResults_cat;
  end
end

local function getAchievementID(name, list, anyCase, allowPartial)
  list = list or CATEGORIES_ALL
  if (type(list) == "table") then
    local id
    for i,cat in ipairs(list) do
      id = getAchievementID_cat(cat, name, anyCase, allowPartial)
      if (id) then  return id;  end
    end
  elseif (type(list) == "number") then
    return getAchievementID_cat(list, name, anyCase, allowPartial)
  elseif (type(list) == "string") then
    local cat = getCategoryID(id)
    assert(cat, "Category not found.")
    return getAchievementID_cat(cat, name, anyCase, allowPartial)
  end
end

local searchResults
local function SearchAchievements(name, list)
  list = list or CATEGORIES_ALL
  if (type(list) == "table") then
    if (searchResults) then
      for k,v in pairs(searchResults) do  -- Reuse table to avoid garbage generation
        searchResults[k] = nil
      end
    else
      searchResults = {}
    end
    local tab
    for i,cat in ipairs(list) do
      tab = getAchievementID_cat(cat, name, true, true, true)
      if (tab) then
        for _,v in ipairs(tab) do
          searchResults[#(searchResults) + 1] = v;
        end
      end
    end
    if (#(searchResults) > 0) then
      return searchResults;
    end
  elseif (type(list) == "number") then
    return getAchievementID_cat(list, name, true, true, true)
  elseif (type(list) == "string") then
    local cat = getCategoryID(id)
    assert(cat, "Category not found.")
    return getAchievementID_cat(cat, name, true, true, true)
  end
end

local function setTracking(id)
  SetTrackedAchievement(id)
  if (AchievementFrameAchievements_ForceUpdate) then
    AchievementFrameAchievements_ForceUpdate()
  end
end

local function isCriteria(achID, name)
  local n, t, complete
  for i=1,GetAchievementNumCriteria(achID) do
    n, t, complete = GetAchievementCriteriaInfo(achID, i)
    if (n == name) then  return true, complete;  end
  end
end

-- DRAGGABLE FRAMES
---------------------

local function CheckDraggable_AchFrame(self, key, val, LoadPos)
  if (AchievementFrame) then
    -- Check if draggable:
    if (Overachiever_Settings.Draggable_AchFrame) then
      if (not MadeDraggable_AchFrame) then
        TjDragIt.EnableDragging(AchievementFrame, AchievementFrameHeader, AchievementFrameCategoriesContainer)
        MadeDraggable_AchFrame = true
      end
    elseif (MadeDraggable_AchFrame) then
      TjDragIt.DisableDragging(AchievementFrame, AchievementFrameHeader, AchievementFrameCategoriesContainer)
      MadeDraggable_AchFrame = nil
    end
    if (key and AchievementFrame:IsShown()) then
      ToggleAchievementFrame() -- If checkbox used, frame needs to be hidden before changing its attributes.
    end
    -- Check if position saved:
    if (Overachiever_Settings.DragSave_AchFrame) then
      if (not MadeDragSave_AchFrame) then
        if (not Overachiever_CharVars.Pos_AchievementFrame) then
          Overachiever_CharVars.Pos_AchievementFrame = {}
        end
        TjDragIt.EnablePositionSaving(AchievementFrame, Overachiever_CharVars.Pos_AchievementFrame, LoadPos)
        AchievementFrame:SetAttribute("UIPanelLayout-enabled", false);
        MadeDragSave_AchFrame = true
      end
    elseif (MadeDragSave_AchFrame) then
      TjDragIt.DisablePositionSaving(AchievementFrame)
      AchievementFrame:SetAttribute("UIPanelLayout-enabled", true);
      MadeDragSave_AchFrame = nil
    end
  end
end

local orig_AchievementFrame_area

local function AchievementUI_Loaded_postPossibleShow()
  Overachiever.MainFrame:Hide()
  -- Delayed setting the area attribute until now so that if Blizzard_AchievementUI was loaded just before the
  -- AchievementFrame is to be shown, AchievementFrame wouldn't be set as the current UI panel by code in
  -- UIParent.lua, which causes problems when we don't want it to interact with other UI panels in the standard
  -- way. (Set now instead of leaving it out because the player may want it to interact normally again later.)
  if (orig_AchievementFrame_area) then
    UIPanelWindows["AchievementFrame"].area = orig_AchievementFrame_area
    AchievementFrame:SetAttribute("UIPanelLayout-area", orig_AchievementFrame_area);
  end
  CheckDraggable_AchFrame(nil, nil, nil, true)
end



-- ACHIEVEMENT HYPERLINK HOOK
-------------------------------

local orig_SetItemRef = SetItemRef
SetItemRef = function(link, text, button, ...)
  if ( strsub(link, 1, 11) == "achievement" and IsControlKeyDown() ) then
    local id = strsplit(":", strsub(link, 13));
    id = tonumber(id)
    openToAchievement(id, true)
    return;
  end
  return orig_SetItemRef(link, text, button, ...)
end

-- ACHIEVEMENT TRACKER CHANGES
--------------------------------

-- local orig_AchievementWatchButton_OnClick = AchievementWatchButton_OnClick

-- We're replacing this function instead of just hooking it because in addition to some new functionality, we prefer
-- an openToAchievement() call instead of a direct AchievementFrame_SelectAchievement() call.
AchievementWatchButton_OnClick = function()
  if (IsShiftKeyDown()) then
    if (not ChatFrameEditBox:IsVisible()) then  ChatFrameEditBox:Show()  end
    local link = GetAchievementLink(AchievementWatchFrame.achievementID)
    ChatEdit_InsertLink(link)
  else
    openToAchievement(AchievementWatchFrame.achievementID)
  end
end

local function TrackerBtnOnEnter(self)
  GameTooltip:SetOwner(self, "ANCHOR_BOTTOMRIGHT");
  local link = GetAchievementLink(AchievementWatchFrame.achievementID)
  GameTooltip:SetHyperlink(link)
end

local function TrackerBtnOnLeave()
  GameTooltip:Hide()
end

-- There are normally no OnEnter or OnLeave scripts for this button, but we check for them in case something else
-- adds one.
if (AchievementWatchLine1Button:GetScript("OnEnter")) then
  AchievementWatchLine1Button:HookScript("OnEnter", TrackerBtnOnEnter)
else
  AchievementWatchLine1Button:SetScript("OnEnter", TrackerBtnOnEnter)
end
if (AchievementWatchLine1Button:GetScript("OnLeave")) then
  AchievementWatchLine1Button:HookScript("OnLeave", TrackerBtnOnLeave)
else
  AchievementWatchLine1Button:SetScript("OnLeave", TrackerBtnOnLeave)
end

local function CreateGreenCheck(refobj)
  local frame = CreateFrame("Frame", refobj)
  frame:SetWidth(20); frame:SetHeight(16);
  frame.tex = frame:CreateTexture(nil, "ARTWORK")
  frame.tex:SetTexture("Interface\\AchievementFrame\\UI-Achievement-Criteria-Check")
  frame.tex:SetTexCoord(0, 0.625, 0, 1)
  frame.tex:SetWidth(20); frame.tex:SetHeight(16);
  frame:SetPoint("RIGHT", refobj, "LEFT", 0, -3)
  frame.tex:SetPoint("LEFT")
  frame:SetScale(0.85)
  frame:SetFrameLevel(refobj:GetFrameLevel() + 1)
  return frame;
end

local function GreenCheckUpdate()
  if (Overachiever_Settings.Tracker_GreenCheck and Overachiever_CharVars.TrackedAch) then
    local _, _, _, complete = GetAchievementInfo(Overachiever_CharVars.TrackedAch)
    if (complete) then
      Overachiever.TrackingGreenCheck = Overachiever.TrackingGreenCheck or CreateGreenCheck(AchievementWatchLine1Button)
      Overachiever.TrackingGreenCheck:Show()
      return;
    end
  end
  if (Overachiever.TrackingGreenCheck and Overachiever.TrackingGreenCheck:IsVisible()) then
    Overachiever.TrackingGreenCheck:Hide()
  end
end

local function saveTrackedAchievement(id, ...)
  if (id == 0) then  id = nil;  end
  Overachiever_CharVars.TrackedAch = id;
  GreenCheckUpdate()
end
hooksecurefunc("SetTrackedAchievement", saveTrackedAchievement)


-- META-CRITERIA TOOLTIP
--------------------------

local orig_AchievementButton_GetMeta

local function MetaCriteriaOnEnter(self)
  if (self.id) then
    GameTooltip:SetOwner(self, "ANCHOR_RIGHT");
    local link = GetAchievementLink(self.id)
    GameTooltip:SetHyperlink(link)
  -- If for some reason the ID isn't there but a date completed is (something that probably shouldn't happen), fall
  -- back to the old way tooltips were handled in Blizzard_AchievementUI.xml's MetaCriteriaTemplate:
  elseif ( self.date ) then
    GameTooltip:SetOwner(self, "ANCHOR_RIGHT");
    GameTooltip:AddLine(self.date, 1, 1, 1);
    GameTooltip:Show();
  end
end

local function new_AchievementButton_GetMeta(...)
  local frame = orig_AchievementButton_GetMeta(...)
  frame:SetScript("OnEnter", MetaCriteriaOnEnter)
  return frame;
end


-- TOOLTIP HOOK
-----------------

-- "To All The Squirrels I've Loved Before" achievement:
local ACH_LoveCritters, ACH_LoveCritters_Complete
local ACH_LoveCritters_ShowComplete = true

local function tooltipSetUnit(tooltip)
  if (ACH_LoveCritters_Complete or not Overachiever_Settings.CritterTip_loved) then  return;  end
  local name, unit = tooltip:GetUnit()
  if (name and unit and UnitCreatureType(unit) == OVERACHIEVER_STRINGS_CRITTER) then
    ACH_LoveCritters = ACH_LoveCritters or getAchievementID(OVERACHIEVER_STRINGS_ACH_LOVECRITTERS)
    local _, _, _, AchComplete = GetAchievementInfo(ACH_LoveCritters)
    if (AchComplete) then
      ACH_LoveCritters_Complete = true;
      return;
    end

    local isCrit, complete = isCriteria(ACH_LoveCritters, name)
    local text, r, g, b
    if (isCrit) then
      if (not complete) then
        text = OVERACHIEVER_STRINGS_ACH_LOVECRITTERS_INCOMPLETE
        r, g, b = 1, 0.1, 0.1
      elseif (ACH_LoveCritters_ShowComplete) then
        text = OVERACHIEVER_STRINGS_ACH_LOVECRITTERS_COMPLETE
        r, g, b = 0.2, 0.5, 0.2
      end
    end
    if (text) then
      tooltip:AddLine(text, r, g, b)
      tooltip:AddTexture(AchievementIcon)
    end
  end
end

GameTooltip:HookScript("OnTooltipSetUnit", tooltipSetUnit)


-- GLOBAL FUNCTIONS
-----------------------

function Overachiever.OnEvent(self, event, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9)
  if (Overachiever_Debug and (arg1 or arg2 or arg3 or arg4 or arg5 or arg6 or arg7 or arg8 or arg9)) then
    print("Overachiever event:", event)
    print(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9)
  end
  if (event == "PLAYER_ENTERING_WORLD") then
    Overachiever.MainFrame:UnregisterEvent("PLAYER_ENTERING_WORLD")
    Overachiever.MainFrame:RegisterEvent("ZONE_CHANGED_NEW_AREA")
    if (not Overachiever_Settings) then
      Overachiever_Settings = Overachiever.DefaultSettings;
    else
      Overachiever_Settings.Version = THIS_VERSION
    end
    local tracked = GetTrackedAchievement()
    if (Overachiever_CharVars) then
      if (not tracked) then  -- Resume tracking from last session.
        tracked = Overachiever_CharVars.TrackedAch
      end
    else
      Overachiever_CharVars = {};
    end
    Overachiever_CharVars.Version = THIS_VERSION
    setTracking( (tracked or 0) )

    OptionsPanel = TjOptions.CreatePanel("Overachiever", nil, {
	title = "Overachiever v"..THIS_VERSION,
	itemspacing = 4,
	items = {
		 { type = "labelnorm", text = OVERACHIEVER_STRINGS_OPT_LABEL_TOOLTIPS, topBuffer = 4 },
		 { variable = "Tooltip_ShowProgress", text = OVERACHIEVER_STRINGS_OPT_SHOWPROGRESS,
		   tooltip = OVERACHIEVER_STRINGS_OPT_SHOWPROGRESS_TIP },
		 { variable = "Tooltip_ShowProgress_Other", text = OVERACHIEVER_STRINGS_OPT_SHOWPROGRESS_OTHER,
		   tooltip = OVERACHIEVER_STRINGS_OPT_SHOWPROGRESS_OTHER_TIP, xOffset = 10 },
		 { variable = "Tooltip_ShowID", text = OVERACHIEVER_STRINGS_OPT_SHOWID, xOffset = 0 },
		 { type = "labelnorm", text = OVERACHIEVER_STRINGS_OPT_LABEL_TRACKING, topBuffer = 8 },
		 { variable = "Tracker_GreenCheck", text = OVERACHIEVER_STRINGS_OPT_TRACKERGREENCHECK,
		   tooltip = OVERACHIEVER_STRINGS_OPT_TRACKERGREENCHECK_TIP, OnChange = GreenCheckUpdate },
		 { variable = "Explore_AutoTrack", text = OVERACHIEVER_STRINGS_OPT_AUTOTRACKEXPLORE,
		   tooltip = OVERACHIEVER_STRINGS_OPT_AUTOTRACKEXPLORE_TIP },
		 { variable = "Explore_AutoTrack_Completed",
		   text = OVERACHIEVER_STRINGS_OPT_AUTOTRACKEXPLORE_COMPLETED, xOffset = 10 },
		 { type = "labelnorm", text = '"'..OVERACHIEVER_STRINGS_ACH_LOVECRITTERS..'"', topBuffer = 8,
		   xOffset = 0 },
		 { variable = "CritterTip_loved", text = OVERACHIEVER_STRINGS_OPT_CRITTERTIPS,
		   tooltip = OVERACHIEVER_STRINGS_OPT_CRITTERTIPS_TIP },
		 { type = "labelnorm", text = OVERACHIEVER_STRINGS_OPT_LABEL_DRAG, topBuffer = 8 },
		 { variable = "Draggable_AchFrame", text = OVERACHIEVER_STRINGS_OPT_DRAG_ACHFRAME,
		   OnChange = CheckDraggable_AchFrame },
		 { variable = "DragSave_AchFrame", text = OVERACHIEVER_STRINGS_OPT_DRAGSAVE, xOffset = 10,
		   OnChange = CheckDraggable_AchFrame },
	},
	variables = Overachiever_Settings,
	defaults = Overachiever.DefaultSettings
    });
  end

  if (event == "ZONE_CHANGED_NEW_AREA" or event == "PLAYER_ENTERING_WORLD") then
    if (Overachiever_Settings.Explore_AutoTrack) then
      local tracked = GetTrackedAchievement()
      -- Don't switch tracked achievement if we're currently tracking a non-Explore achievement.
      if (not tracked or isExplorationAchievement(tracked, true)) then
        local id
        if (not IsInInstance()) then
          --local achName = OVERACHIEVER_STRINGS_EXPLOREZONE:format(GetRealZoneText())
          id = getAchievementID(GetRealZoneText(), CATEGORIES_EXPLOREZONES, true, true)
        end
        if (not id) then
          if (event ~= "PLAYER_ENTERING_WORLD") then  setTracking(0);  end
        elseif (id ~= tracked) then
          local _, _, _, complete = GetAchievementInfo(id)
          if (not complete or Overachiever_Settings.Explore_AutoTrack_Completed) then
            setTracking(id)
          elseif (complete and event ~= "PLAYER_ENTERING_WORLD") then
            setTracking(0)
          end
        end
      end
    end

  elseif (event == "ACHIEVEMENT_EARNED" and arg1 == Overachiever_CharVars.TrackedAch) then
    GreenCheckUpdate()

  elseif (event == "ADDON_LOADED" and arg1 == "Blizzard_AchievementUI") then
    Overachiever.MainFrame:UnregisterEvent("ADDON_LOADED")
    -- META-CRITERIA CREATION INTERCEPT:
    orig_AchievementButton_GetMeta = AchievementButton_GetMeta
    AchievementButton_GetMeta = new_AchievementButton_GetMeta
    -- Allow closing frame with Escape even when UIPanelLayout-enabled is set to false:
    tinsert(UISpecialFrames, "AchievementFrame");
    -- Make main achievement UI draggable:
    if (Overachiever_Settings.DragSave_AchFrame) then
      orig_AchievementFrame_area = UIPanelWindows["AchievementFrame"].area
      UIPanelWindows["AchievementFrame"].area = nil
    end
    Overachiever.MainFrame:Show()  -- delayed call of AchievementUI_Loaded_postPossibleShow()
  end
end

Overachiever.OpenToAchievement = openToAchievement;

function Overachiever.SearchForAchievement(msg, silent, givelist)
  local id = getAchievementID(msg, nil, true, true)
  if (id) then
    if (not silent) then  chatprint(OVERACHIEVER_STRINGS_MSG_OPENINGTO..GetAchievementLink(id));  end
    openToAchievement(id)
    if (not silent or givelist) then
      local tab = SearchAchievements(msg)
      local tab2
      if (givelist and tab) then
        tab2 = {}
        copytab(tab, tab2)
      end
      if (silent) then  return tab2 or id;  end
      if (tab) then
        local size = #(tab)
        if (size == 2) then
          chatprint(OVERACHIEVER_STRINGS_MSG_ONEFOUND..GetAchievementLink(tab[2]))
        elseif (size > 2) then
          chatprint(OVERACHIEVER_STRINGS_MSG_NUMFOUNDLIST:format(size-1))
          local a, b, c
          for i=2,size,3 do
            a, b, c = tab[i], tab[i+1], tab[i+2]
            a, b, c = GetAchievementLink(a), b and GetAchievementLink(b), c and GetAchievementLink(c)
            if (b) then  a = a.."  --  "..b;  end
            if (c) then  a = a.."  --  "..c;  end
            chatprint(a, "-- ")
          end
        end
      end
      return tab2 or id;
    else
      return id;
    end
  elseif (not silent) then
    chatprint(OVERACHIEVER_STRINGS_MSG_NAMENOTFOUND:format(msg))
  end
end


-- SLASH COMMANDS
-------------------

function Overachiever.SlashHandler(msg, self, silent)
  if (msg == "") then
    ToggleAchievementFrame();
  else
    if (strsub(msg, 1,1) == "#") then
      local id = tonumber(strsub(msg, 2))
      if (id) then
        if (GetAchievementInfo(id)) then
          if (not silent) then  chatprint(OVERACHIEVER_STRINGS_MSG_OPENINGTO..GetAchievementLink(id));  end
          openToAchievement(id)
        elseif (not silent) then
          chatprint(OVERACHIEVER_STRINGS_MSG_INVALIDID);
        end
        return;
      end
    end
    Overachiever.SearchForAchievement(msg, silent)
  end
end

SLASH_Overachiever1 = "/oa";
SlashCmdList["Overachiever"] = Overachiever.SlashHandler;
-- Replace the original slash command handler for /ach, /achieve, /achievement, and /achievements:
SlashCmdList["ACHIEVEMENTUI"] = Overachiever.SlashHandler;

SLASH_Overachiever_silent1 = "/oasilent";
SLASH_Overachiever_silent2 = "/oas";
SLASH_Overachiever_silent3 = "/achsilent";
SLASH_Overachiever_silent4 = "/achs";
SlashCmdList["Overachiever_silent"] = function(msg, self)  Overachiever.SlashHandler(msg, self, true);  end;


-- LOCALIZATION ASSIST
--------------------------

local ZoneID

local function BuildZoneIDTable_z(connum, num, zonename, next, ...)
  num = num + 1
  ZoneID[zonename] = {}
  ZoneID[zonename].C = connum
  ZoneID[zonename].Z = num
  if (not ZoneDefense_subzones[zonename]) then
    ZoneDefense_subzones[zonename] = zonename;
    --print('Added: ZoneDefense_subzones["'..zonename..'"] = "'..zonename..'"')
  elseif (ZoneDefense_subzones[zonename] ~= zonename) then
    print('Problem with data: ZoneDefense_subzones["'..zonename..'"] ~= "'..zonename..'" !')
  end
  if (next) then
    BuildZoneIDTable_z(connum, num, next, ...)
  end
end

-- Intended first call:  BuildZoneIDTable(0, GetMapContinents());
local function BuildZoneIDTable(num, conname, next, ...)
  num = num + 1
  --print("BuildZoneIDTable: "..conname.." ("..num..")")
  BuildZoneIDTable_z(num, 0, GetMapZones(num))
  if (next) then
    BuildZoneIDTable(num, next, ...)
  end
end

function Overachiever.Debug_GetExplorationData()
  if (not ZoneID) then
    ZoneID = {}
    BuildZoneIDTable(0, GetMapContinents());
    Overachiever_Settings.Debug_ZoneData = ZoneID
  end
  chatprint("Zone data gathered.")

  assert(CATEGORIES_EXPLOREZONES, "Exploration categories not defined. Check that OVERACHIEVER_STRINGS_CAT_EXPLORATION (currently '"..(OVERACHIEVER_STRINGS_CAT_EXPLORATION or "nil").."') was set properly.")
  Overachiever_Settings.Debug_ExplorationData = {}
  local tab = Overachiever_Settings.Debug_ExplorationData
  local catname
  for _, category in ipairs(CATEGORIES_EXPLOREZONES) do
    catname = GetCategoryInfo(category)
    tab[catname] = {}
    for i=1,GetCategoryNumAchievements(category) do
      tab[catname][i] = { GetAchievementInfo(category, i) }
    end
  end
  chatprint("Exploration Achievement data gathered.")

  chatprint("Reload the interface or log out to output gathered data to saved variables file.")
end


-- FRAME INITIALIZATION
--------------------------

Overachiever.MainFrame = CreateFrame("Frame")
Overachiever.MainFrame:Hide()
Overachiever.MainFrame:RegisterEvent("PLAYER_ENTERING_WORLD")
Overachiever.MainFrame:RegisterEvent("ADDON_LOADED")
Overachiever.MainFrame:RegisterEvent("ACHIEVEMENT_EARNED")
if (Overachiever_Debug) then
  Overachiever.MainFrame:RegisterEvent("CRITERIA_UPDATE")
end
Overachiever.MainFrame:SetScript("OnEvent", Overachiever.OnEvent)
Overachiever.MainFrame:SetScript("OnUpdate", AchievementUI_Loaded_postPossibleShow)

BuildCategoryInfo();
BuildCategoryInfo = nil;
